// DigitalInput.js
//  This module handles digital input processing and eventing

//#region Dependencies, Constants and Globals
//#region Dependencies
var fs = require('fs');
var moment = require('moment');
var exec = require('child_process').exec;
//#endregion Dependencies

//#region Constants
// Define the configuraton for the I/O
var DIGITAL_OUTPUTS_LOCATION = '/sys/class/gpio/gpio';
var DIGITAL_OUTPUTS_BASEGPIO = 496;
var DIGITAL_OUTPUTS_POINTS = 16;// Default = 16

// SNMP - this is here for now, could move
var AuthProtocol = ["None", "MD5", "SHA"];
var PrivProtocol = ["None", "DES", "AES"];
//#endregion Constants

//#region Globals
var digitalOutputsArray = [
    { "Address" : "0", "Name" : "Name", "Value": 0 , "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "ActionType": 0 , "TimeChanged": 0 }, 
    { "Address" : "1", "Name" : "Name", "Value": 0 , "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "ActionType": 0 , "TimeChanged": 0 }, 
    { "Address" : "2", "Name" : "Name", "Value": 0 , "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "ActionType": 0 , "TimeChanged": 0 }, 
    { "Address" : "3", "Name" : "Name", "Value": 0 , "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "ActionType": 0 , "TimeChanged": 0 }, 
    { "Address" : "4", "Name" : "Name", "Value": 0 , "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "ActionType": 0 , "TimeChanged": 0 },
    { "Address" : "5", "Name" : "Name", "Value": 0 , "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "ActionType": 0 , "TimeChanged": 0 }, 
    { "Address" : "6", "Name" : "Name", "Value": 0 , "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "ActionType": 0 , "TimeChanged": 0 }, 
    { "Address" : "7", "Name" : "Name", "Value": 0 , "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "ActionType": 0 , "TimeChanged": 0 }, 

    { "Address" : "8", "Name" : "Name", "Value": 0 , "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "ActionType": 0 , "TimeChanged": 0 }, 
    { "Address" : "9", "Name" : "Name", "Value": 0 , "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "ActionType": 0 , "TimeChanged": 0 }, 
    { "Address" : "10", "Name" : "Name", "Value": 0 , "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "ActionType": 0 , "TimeChanged": 0 }, 
    { "Address" : "11", "Name" : "Name", "Value": 0 , "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "ActionType": 0 , "TimeChanged": 0 }, 
    { "Address" : "12", "Name" : "Name", "Value": 0 , "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "ActionType": 0 , "TimeChanged": 0 }, 
    { "Address" : "13", "Name" : "Name", "Value": 0 , "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "ActionType": 0 , "TimeChanged": 0 }, 
    { "Address" : "14", "Name" : "Name", "Value": 0 , "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "ActionType": 0 , "TimeChanged": 0 }, 
    { "Address" : "15", "Name" : "Name", "Value": 0 , "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "ActionType": 0 , "TimeChanged": 0 }
];

var SummitInfo = {};
var SmtpManager = {};
var SnmpManager = {};
var SqliteManager = {};
var LogManager = {};
var SocketIoManager = {};
//#endregion Globals
//#endregion Dependencies, Constants and Globals

//#region Digital Outputs Initialization
// Initialize the analog outputs upon system start
//      - Checks if database table exists, if not creates the table and fills it with default data for each point
//      - Else reads from the database and updates digitalOutputsArray
exports.InitializeDigitalOutputs = function (info, smtp, snmp, db, io, logger, callback) {
    // Initialize globals
    SummitInfo = info;
    SmtpManager = smtp;
    SnmpManager = snmp;
    SqliteManager = db;
    SocketIoManager = io;
    LogManager = logger;
    
    // Try getting the DigitalOutputs table
    SqliteManager.get("SELECT name FROM sqlite_master WHERE type='table' AND name='DigitalOutputs'", function (err, rows) {
        if (err !== null) {
            LogManager.error('Digital Outputs - Initialize Error (Check)=' + JSON.stringify(err));
        }
        else if (typeof rows === "undefined") {
            // We read nothing, so let's create the table ...
            LogManager.info("Digital Outputs - Initialize by creating database table ...");
            SqliteManager.exec("CREATE TABLE if not exists DigitalOutputs (Address INTEGER, Name TEXT, Value TEXT, Enabled TEXT, SendEvents TEXT, Polarity TEXT, ActionType TEXT, TimeChanged TEXT )", function (err) {
                if (err !== null) {
                    LogManager.error('Digital Outputs - Initialize Error (Create)=' + JSON.stringify(err));
                }
                else {
                    var timeChanged = moment();// Record the create time as the last time the I/O point value was changed
                    
                    // Created the table, now add data ...
                    SqliteManager.run("BEGIN")
                    for (var i = 0; i < DIGITAL_OUTPUTS_POINTS; i++) {
                        // Initialize the array
                        UpdateDigitalOutputConfiguration(i, "Digital Output " + (i + 1), 1, "FALSE", "FALSE", "Normally Open", "Latched", timeChanged);
                        // Empty space here to be consistent
                        // Empty space here to be consistent
                        
                        // Initialize the database
                        insertStatement = i + ", 'Digital Output " + (i + 1) + "', '1', 'FALSE', 'FALSE', 'Normally Open', 'Latched', '" + timeChanged + "'";
                        SqliteManager.run("INSERT INTO DigitalOutputs (Address, Name, Value, Enabled, SendEvents, Polarity, ActionType, TimeChanged) VALUES (" + insertStatement + ")", function (err) {
                            if (err !== null) {
                                LogManager.error('Digital Outputs - Initialize Error (Insert)=' + JSON.stringify(err));
                            }
                        });
                    }
                    SqliteManager.run("COMMIT");
                    
                    // Digital Outputs table created
                    if (typeof callback == "function") {
                        LogManager.info('Digital Outputs - Initialized new database table');
                        callback();
                    }
                }
            });
        }
        else {
            // Read from the database and populate the digitalOutputsArray
            LogManager.info("Digital Outputs - Initialize by reading database ...");
            SqliteManager.each("SELECT rowid as id, Address, Name, Value, Enabled, SendEvents, Polarity, ActionType, TimeChanged FROM DigitalOutputs ORDER BY Address", function (err, row) {
                if (err !== null) {
                    LogManager.error('Digital Outputs - Initialize Error (Select)=' + JSON.stringify(err));
                }
                else {
                    UpdateDigitalOutputConfiguration(row.Address, row.Name, row.Value, row.Enabled, row.SendEvents, row.Polarity, row.ActionType, row.TimeChanged);
                }
            }, function () {                
                // Digital Outputs read from table
                if (typeof callback == "function") {
                    LogManager.info('Digital Outputs - Initialized from existing database table');
                    callback();
                }            
            });
        }
    });
};

// Added this in later to support Apex4 Integration
var Apex4Token = null;
var Apex4Manager = null;
exports.ConfigureApex4TokenAndSocket = function (apex4Token, apex4Socket) {
    Apex4Token = apex4Token;
    Apex4Manager = apex4Socket;
}
//#endregion Digital Outputs Initialization

//#region Digital Outputs API
// Adjust the digitalOutputsArray
module.exports.UpdateDigitalOutputConfiguration = UpdateDigitalOutputConfiguration;
function UpdateDigitalOutputConfiguration(Address, Name, Value, Enabled, SendEvents, Polarity, ActionType, TimeChanged) {
    digitalOutputsArray[Address].Name = Name;
    digitalOutputsArray[Address].Value = Value;
    digitalOutputsArray[Address].Enabled = Enabled;
    digitalOutputsArray[Address].SendEvents = SendEvents;
    digitalOutputsArray[Address].Polarity = Polarity;
    digitalOutputsArray[Address].ActionType = ActionType;
    digitalOutputsArray[Address].TimeChanged = TimeChanged;
    sendValueDigitalOutputs();// Update the Digital Outputs configuration
};

// Reset the digital outputs on start-up to ensure a proper value
exports.ResetDigitalOutputs = function (ledManager, callback) {
    // Get the stored Polarity from the database,
    //  then adjust the reset point based on the polarity        
    var timeChanged = moment();
    var digitalOutputs = [];
    SqliteManager.all("SELECT Address, Name, Value, Enabled, SendEvents, Polarity, ActionType, TimeChanged FROM digitaloutputs ORDER BY Address", function (err, rows) {
        if (err) {
            throw LogManager.error('Digital Outputs - Reset Error (Select)=' + JSON.stringify(err));
        }

        // For each point, reset the value based on polarity setting 
        for (index = 0; index < rows.length; index++) {

            if (rows[index].Polarity === "Normally Closed") {
                polarityValue = "1";
            } else {
                polarityValue = "0";
            }

            // Reset the output point ...        
            fs.writeFile(DIGITAL_OUTPUTS_LOCATION + (DIGITAL_OUTPUTS_BASEGPIO + index) + "/value", polarityValue, function (err) {
                if (err) {
                    return LogManager.error('Digital Outputs - Reset Error (Write)=' + JSON.stringify(err));
                }
            });
        }

        // Digital Outputs reset is complete
        if (typeof callback == "function") {
            LogManager.debug('Digital Outputs - Reset to initial values');
            callback();
        }
    });
}

// Set value of a digital output
exports.SetDigitalOutput = function (outputPoint, setValue, callback) {
    LogManager.info('Digital Outputs.SetDigitalOutput');

    // Get the digital output settings to determine if this is NC or NO
    SqliteManager.get("SELECT Polarity FROM digitaloutputs WHERE Address='" + outputPoint + "'", function (err, row) {
        if (typeof row !== "undefined") {
            // Ensure we know the proper polarity
            polarity = row.Polarity;
        }
        LogManager.info("Digital Outputs.SetDigitalOutput: Digital Output #" + (parseInt(outputPoint) + 1) + " Set Value=" + setValue + " Polarity=" + row.Polarity);

        // Determine the value to set
        var index = DIGITAL_OUTPUTS_BASEGPIO + parseInt(outputPoint);
        if (polarity === "Normally Open") {
            if (setValue == "0") {
                value = "0";
            }
            else {
                value = "1";
            }
        }
        else {
            if (setValue == "0") {
                value = "1";
            }
            else {
                value = "0";
            }
        }

        // Set the value ...
        var index = DIGITAL_OUTPUTS_BASEGPIO + parseInt(outputPoint);
        var result;
        fs.writeFile("/sys/class/gpio/gpio" + index + "/value", value, function (err) {
            if (err) {
                // Record the error
                result = JSON.stringify(err);
                LogManager.error("Digital Outputs.SetDigitalOutput: Digital Output #" + (parseInt(outputPoint) + 1) + " Set Value=" + setValue + " Result=" + result);

                // Return with error
                if (typeof callback == "function") {
                    callback(result);
                }
                return;
            }

            // Record the result
            result = 'OK';
            LogManager.debug('Digital Outputs.SetDigitalOutput: Digital Output #' + (parseInt(outputPoint) + 1) + ' Set Value=' + setValue + ' Result=' + result);

            // Return with OK !
            if (typeof callback == "function") {
                callback(result);
            }
        });
    });
}

// Get DigitalOutputsArray
module.exports.GetDigitalOutputsArray = GetDigitalOutputsArray;
function GetDigitalOutputsArray() {
    return digitalOutputsArray;
}
//#endregion Digital Outputs API

//#region Digital Outputs Processing
// Intervalled processing of digital outputs
var processCounter = 0;
var processTotalTime = 0;
exports.ProcessDigitalOutputs = function (ledManager) {
    timeStart = moment();
    
    var index = 0;// This is an index to digital I/O point - accounts for first 24 bits being at a higher location that the second 24 bits
    var fileIndex = 0;// This is an index to the GPIO location - it's the index + the GPIO offset
    var timeChanged = moment();
    for (index = 0; index < DIGITAL_OUTPUTS_POINTS; index++) {
        fileIndex = DIGITAL_OUTPUTS_BASEGPIO + index;
        
        // Only read value if the I/O point is enabled
        if (digitalOutputsArray[index].Enabled === "TRUE") {
            try {
                data = fs.readFileSync(DIGITAL_OUTPUTS_LOCATION + fileIndex + "/value");
                readValue = String(data).replace(/(\r\n|\n|\r)/gm, "");
                
                // Normally Closed vs Normally Open
                if (digitalOutputsArray[index].Polarity == "Normally Closed") {
                    if (readValue === "1") {
                        readValue = "0";
                    } else {
                        readValue = "1";
                    }
                }

                // Check to see if value has changed, if so update database and clients ...
                if (digitalOutputsArray[index].Value != readValue) {
                    digitalOutputsArray[index].Value = readValue;
                    digitalOutputsArray[index].TimeChanged = timeChanged;
                    sendSocketSnmpSmtp(digitalOutputsArray[index])

                    // Value has changed, update the database and then the clients ...
                    SqliteManager.run("UPDATE DigitalOutputs SET Value='" + readValue 
                        + "', TimeChanged='" + timeChanged 
                        + "' WHERE Address = '" + index + "'", 
                        function () {
                    });
                }
                //logger.info("Digital Output[" + (index+1) + "]=" + readValue + " (Enabled)")
            }
            catch (error) {
                LogManager.error('Digital Outputs - Process Error (Read)=' + JSON.stringify(error));
            }
        }
        else {
            //logger.info("Digital Output[" + (index+1) + "]=" + digitalOutputsArray[index].Value + " (Disabled)")
        }
    }
    //if (processCounter === 500) {
    //    logger.info("ProcessDigitalOutputs: Average process time = " + (processTotalTime / processCounter).toFixed(2) + "msec");
    //    processCounter = 0;
    //    processTotalTime = 0;
    //}
    //processCounter++;
    //processTotalTime += (Date.now() - timeStart);
}
//#endregion Digital Outputs Processing

//#region Digital Outputs Eventing
function sendSocketSnmpSmtp(digitalOutput) {
    LogManager.info("Digital Outputs - [Value UPDATED] Digital Output[" + (parseInt(digitalOutput.Address) + 1) + "] '" + digitalOutput.Name + "' Value=" + digitalOutput.Value);
    LogManager.debug("Digital Outputs - [Value UPDATED] Digital Output[" + (parseInt(digitalOutput.Address) + 1) + "] =" + JSON.stringify(digitalOutput));
    
    // SOCKET.IO: Send changed value to all attached clients
    sendValueDigitalOutput("DigitalOutput", digitalOutput);
    
    // SNMP: Send SNMP Trap/Notification if threshold(s) crossed
    sendSNMP(digitalOutput);

    // SMTP: Send SMTP message if threshold(s) crossed
    sendSMTP(digitalOutput)

    // APEX4: Send message to Apex 4
    if (Apex4Manager) {
        var changestate = {};
        var uid = "DO" + digitalOutput.Address;
        changestate["token"] = Apex4Token;
        changestate["time"] = digitalOutput.TimeChanged;
        changestate[uid] = digitalOutput.Value;
        Apex4Manager.emit("changestate", changestate);
        LogManager.info("APEX4 message=" + JSON.stringify(changestate));
    }
}

function sendValueDigitalOutput(dataType, data) {
    SocketIoManager.sockets.emit(dataType, data);
}

function sendValueDigitalOutputs() {
    enabled = 0;
    for (i = 0; i < digitalOutputsArray.length; i++) {
        if (digitalOutputsArray[i].Enabled === "TRUE") enabled++;
    }
    SocketIoManager.sockets.emit("DigitalOutputs", { "DigitalOutputsTotal": digitalOutputsArray.length, "DigitalOutputsEnabled": enabled })
}

function sendSNMP(digitalOutput) {
    // SNMP definitions - also refer to the TASC-SUMMIT.MIB file for detailed information
    if (digitalOutput.Enabled == "FALSE") return;
    if (digitalOutput.SendEvents == "FALSE") return;
    
    address = parseInt(digitalOutput.Address) + 1;// To overcome concatenation
    LogManager.trace("Digital Outputs - Sending SNMP Trap for Digital Output[" + address + "] + '" + digitalOutput.Name + "'");

    // Define the notification
    var snmpCommand;
    var snmpHost;
    var snmpNotification = "";
    snmpNotification += "1.3.6.1.4.1.11902.3.2.1000.1000.0." + (address - 1 + 1000) + " ";
    snmpNotification += "1.3.6.1.4.1.11902.3.2." + (address) + ".0.1 s '" + moment().toString() + "' ";
    snmpNotification += "1.3.6.1.4.1.11902.3.2." + (address) + ".0.2 i " + (parseInt(digitalOutput.Address) + 1) + " ";
    snmpNotification += "1.3.6.1.4.1.11902.3.2." + (address) + ".0.3 s '" + digitalOutput.Name + "' ";
    snmpNotification += "1.3.6.1.4.1.11902.3.2." + (address) + ".0.4 s '" + ((digitalOutput.Value === "1") ? "ON" : "OFF") + "' ";
    snmpNotification += "1.3.6.1.4.1.11902.3.2." + (address) + ".0.5 i " + ((digitalOutput.Value === "1") ? 1 : 0) + " ";
    snmpNotification += "1.3.6.1.4.1.11902.3.2." + (address) + ".0.6 i " + ((digitalOutput.Polarity === "Normally Closed") ? 1 : 0) + " ";
    snmpNotification += "1.3.6.1.4.1.11902.3.2." + (address) + ".0.20 s '" + SummitInfo.MACID + "' ";
    snmpNotification += "1.3.6.1.4.1.11902.3.2." + (address) + ".0.21 s '" + SummitInfo.Name + "' ";

    // Send the SNMP notification
    for (var i = 0; i < SnmpManager.length; i++) {
        // Skip if this host is not enabled
        if (SnmpManager[i].Enabled === "FALSE") continue;

        // Set up the SNMP trap command
        snmpCommand = "snmptrap ";
        if (SnmpManager[i].SnmpVersion == 2) {
            snmpCommand += "-v 2c ";
            snmpCommand += "-c " + SnmpManager[i].Community + " ";
        }
        else if (SnmpManager[i].SnmpVersion == 3) {
            snmpCommand += "-v 3 ";
            var EngineId = "0x" + SnmpManager[i].Community.replace(/\-/g, '');
            snmpCommand += "-e " + EngineId + " ";
            snmpCommand += "-u " + SnmpManager[i].AuthUser + " ";
            if (SnmpManager[i].AuthProtocol > 0) snmpCommand += "-a " + AuthProtocol[SnmpManager[i].AuthProtocol] + " -A " + SnmpManager[i].AuthPassword + " ";
            if (SnmpManager[i].PrivProtocol > 0) snmpCommand += "-x " + PrivProtocol[SnmpManager[i].PrivProtocol] + " -X " + SnmpManager[i].PrivPassword + " ";
            var securityLevel = "-l noAuthNoPriv ";
            if (SnmpManager[i].AuthProtocol > 0 && SnmpManager[i].PrivProtocol == 0)
                securityLevel = "-l authNoPriv ";
            if (SnmpManager[i].AuthProtocol > 0 && SnmpManager[i].PrivProtocol > 0)
                securityLevel = "-l authPriv ";
            snmpCommand += securityLevel;
        } else continue;

        // Send trap to host ...
        snmpHost = SnmpManager[i].IpAddress + " '' ";
        LogManager.info("Digital Outputs - Sending SNMP Event to '" + SnmpManager[i].HostName + "' for Digital Output[" + address + "] '" + digitalOutput.Name + "' Value=" + ((digitalOutput.Value === "1") ? "ON" : "OFF"));
        LogManager.trace("Digital Outputs - SendSNMP: " + snmpCommand + snmpHost + snmpNotification);
        exec(snmpCommand + snmpHost + snmpNotification, function (error, stdout, stderr) {
            if (error !== null) {
                LogManager.error("Digital Outputs - SendSNMP. " + error);
            } else {
                LogManager.debug("Digital Outputs - SendSNMP. " + stdout);
            }
        });
    }
}

function sendSMTP(digitalOutput) {
    // SNMP definitions - also refer to the TASC-SUMMIT.MIB file for detailed information
    if (digitalOutput.Enabled == "FALSE") return;
    if (digitalOutput.SendEvents == "FALSE") return;
    
    // Set up the SMTP Email message
    var message = [];
    address = parseInt(digitalOutput.Address) + 1;// To overcome concatenation
    var event = "'" + digitalOutput.Name + "' has changed to status=" + ((digitalOutput.Value === "1") ? "ON":"OFF");
    message += moment().toString() + "\n";
    message += "Summit MACID=" + SummitInfo.MACID + "\n";
    message += "Summit Name=" + SummitInfo.Name + "\n";
    message += "Address=" + (parseInt(digitalOutput.Address) + 1) + "\n";
    message += "Name=" + digitalOutput.Name + "\n";
    message += "Value=" + ((digitalOutput.Value === "1") ? "ON":"OFF") + "\n";
    message += "Status=" + ((digitalOutput.Value === "1") ? 1:0) + "\n";
    message += "Polarity=" + ((digitalOutput.Polarity === "Normally Closed") ? 1: 0) + "\n";
    
    var ioInfo = "Digital Output[" + address + "] '" + digitalOutput.Name + "'";
    
    LogManager.info("Digital Outputs - Sending SMTP Event for Digital Outputs[" + address + "] '" + digitalOutput.Name + "' Value=" + ((digitalOutput.Value === "1") ? "ON":"OFF"));
    SmtpManager.SendSmtp(message, ioInfo, event, function (result) {
        LogManager.trace("Digital Outputs - SMTP Email Event Response for Digital Output[" + address + "] '" + digitalOutput.Name + "' Value=" + ((digitalOutput.Value === "1") ? "ON":"OFF") + ": " + result.result);
    })
}


exports.UpdateSnmpManager = function (snmp) {
    SnmpManager = snmp;
}

exports.UpdateSummitInfo = function (info) {
    SummitInfo = info;
}
//#endregion Digital Outputs Eventing